import argparse
import re
from typing import List

from smvp.utilities import print_docstring
from smvp.utilities import task_runner
from smvp.version import get_version


def font_size(size: str) -> str:
    """Validate size inputs.

    Parameters
    ----------
    size : str
        User input for a font size option.

    Returns
    -------
    str
        The validated user input.

    Raises
    ------
    argparse.ArgumentTypeError
        If the input is not a valid integer (for example is a float).
    argparse.ArgumentTypeError
        If the input is not between 2 and 100.
    """
    for c in size:
        if not c.isdigit():
            msg = "Size must be a valid integer"
            raise argparse.ArgumentTypeError(msg)

    int_size = int(size)
    min_size = 2
    max_size = 100
    if int_size >= min_size and int_size <= max_size:
        return size
    else:
        msg = f"Font size must be between {min_size} and {max_size}"
        raise argparse.ArgumentTypeError(msg)


# ======================================================================


def font_family(font: str) -> str:
    """Validate font_family inputs.

    Parameters
    ----------
    font : str
        User input for a font family.

    Returns
    -------
    str
        The validated user input.

    Raises
    ------
    argparse.ArgumentTypeError
        If the the selected font family is not recognized.
    """
    valid_fonts = {
        "ANDALE MONO": "Andale Mono",
        "ARIAL": "Arial",
        "BRUSH SCRIPT MT": "Brush Script MT",
        "COMIC SANS MS": "Comic Sans MS",
        "COURIER NEW": "Courier New",
        "FANTASY": "fantasy",
        "GARAMOND": "Garamond",
        "GEORGIA": "Georgia",
        "HELVETICA": "Helvetica",
        "IMPACT": "Impact",
        "LUMINARI": "Luminari",
        "MONACO": "Monaco",
        "MONOSPACE": "monospace",
        "SANS-SERIF": "sans-serif",
        "SERIF": "serif",
        "TAHOMA": "Tahoma",
        "TIMES NEW ROMAN": "Times New Roman",
        "TREBUCHET MS": "Trebuchet MS",
        "VERDANA": "Verdana",
    }
    user_input = " ".join([word.upper() for word in font.split()])
    if user_input in valid_fonts:
        return valid_fonts[user_input]
    else:
        print()
        fonts = [f'"{token}"' for token in list(valid_fonts.values())]
        fonts.sort()
        msg = f"""
        The font you entered ({font}) is not valid. The default font is
        "Courier New". If you're changing the default font, please use
        one of the options below. Check the spelling to make sure it's
        correct.
        """
        print_docstring(msg=msg)
        print()
        chunk_size = 5
        start = 0
        end = 4
        chunks: List[str] = []
        for font in fonts:
            chunk = fonts[start:end]
            if start > len(fonts) or len(chunk) == 0:
                break
            if len(chunk) > 1:
                chunks.append(", ".join(fonts[start:end]))
            else:
                chunks.append(chunk[0])
            start = end
            end += chunk_size
        print(",\n".join(chunks))
        print()
        raise argparse.ArgumentTypeError("invalid font family")


# ======================================================================


def email_type(address: str) -> str:
    """Validate user input of email addresses.

    Parameters
    ----------
    address : str
        An email address.

    Returns
    -------
    str
        If the address is valid, it's returned.

    Raises
    ------
    argparse.ArgumentTypeError
        This is raised for an invalid email address.
    """
    S = r"[a-zA-Z"
    email = re.compile(rf"^{S}0-9._%+-]+@{S}0-9.-]+\.{S}]{{2,}}$")
    if not email.match(address):
        raise argparse.ArgumentTypeError(f"'{address}' is not a valid email address")
    return address


# ======================================================================


def process_args() -> None:
    msg = """
    Send Mail Via Python (smvp). This tool will send an email from the
    command line, with the body of the email taken from a specified
    file. For example, it's handy to use smvp to have automated Linux
    scripts (i.e. cron jobs) email you status updates and the contents
    of log files.
    """
    epi = f"Version: {get_version()}"
    parser = argparse.ArgumentParser(description=msg, epilog=epi)

    msg = """The email address of the recipient."""
    parser.add_argument("recipient", type=email_type, help=msg)

    msg = """
    The subject of the email. IMPORTANT: Make sure to enclose the entire
    subject in double quotes for proper processing on the command line.
    """
    parser.add_argument("subject", type=str, help=msg)

    msg = """
    The file containing the text which will make up the body of the
    email message. The input file can be a text file with ANSI color
    codes, HTML, or plain text. The resulting email will be sent as a
    multi-part MIME message that renders properly in both plain text and
    HTML.
    """
    parser.add_argument("file", type=argparse.FileType("r"), help=msg)

    msg = """
    Enter the desired font family (enclosed in quotes). Values here are
    not case sensitive. See the README.md file for available options.
    Default = \"Courier New\".
    """
    parser.add_argument(
        "-f", "--font_family", type=font_family, default="Courier New", help=msg
    )

    msg = """
    Enter the desired font pixel size as an integer. Valid sizes are
    between 2 and 100. Default = 12.
    """
    parser.add_argument("-s", "--font_size", type=font_size, default=12, help=msg)

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"smvp {get_version()}",
    )

    args = parser.parse_args()
    task_runner(args=args)
    return



import argparse
import os
import re
import smtplib
import ssl
import sys
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from ansi2html import Ansi2HTMLConverter
from bs4 import BeautifulSoup
from bs4.element import Tag

STYLE = "font-family: FF !important; font-size: FSpx !important;"


def print_docstring(msg: str) -> None:
    """Print a formatted docstring.

    This function assumes the docstring is in a very specific format:

    >>> msg = \"\"\"
    >>> First line (non-blank)
    >>>
    >>> Subsequent lines
    >>> Subsequent lines
    >>> Subsequent lines
    >>> ...
    >>> Can include empty lines after the first.
    >>> \"\"\"

    Parameters
    ----------
    msg : str
        The docstring to be printed.
    """
    # Delete the first line ('\n' by itself), then strip any trailing
    # spaces. Remove leading padding, then print.
    clean = msg[1:].rstrip()
    lines = clean.split("\n")
    spaces = 0
    for c in lines[0]:
        if c.isspace():
            spaces += 1
        else:
            break
    formatted_docstring = "\n".join([line[spaces:] for line in lines])
    print(formatted_docstring)
    return


# ======================================================================


def file_is_html(file_to_test: str) -> bool:
    """Determine if a file is HTML

    Parameters
    ----------
    file_to_test : str
        The text of a file to test.

    Returns
    -------
    bool
        True if the file is HTML; else False
    """
    if re.search(
        r"<!DOCTYPE\s+html>|<(html|head|body|title|div|p|span)",
        file_to_test,
        re.IGNORECASE,
    ):
        return True
    return False


# ======================================================================


def validate_environment() -> bool:
    """Ensure the correct environment variables are in place.

    Returns
    -------
    bool
        True if the correct variables are set; else False.
    """
    try:
        os.environ["SMVP_USER"]
        os.environ["SMVP_TOKEN"]
        os.environ["SMVP_SERVER"]

    except KeyError:
        msg = """
        One or more credentials for sending email are missing from your
        environment. Make sure the following environment variables are
        set and exported in your current shell:
        
        export SMVP_USER="<your email>"    # e.g. "myemail@gmail.com"
        export SMVP_TOKEN="<your token>"   # e.g. "<gmail app password>"
        export SMVP_SERVER="<smtp server>" # e.g. "smtp.gmail.com"

        It's recommended that you put the lines above in your "rc" file
        (.bashrc, .zshrc, etc.) for use across multiple shell sessions
        and processes. To confirm you have the environment variables
        correctly set (with the correct spellings), run this in a
        terminal:

        set | grep ^SMVP_

        Note: If you make changes to your "rc" file, make sure to
        "source" it before running smvp again. Also, the SMVP_SERVER you
        select must support secure TLS connections on port 587.
        """
        print()
        print_docstring(msg=msg)
        return False

    return True


# ======================================================================


def task_runner(args: argparse.Namespace) -> None:
    """Package email contents and send message

    Parameters
    ----------
    args : argparse.Namespace
        The collection of command line arguments.
    """
    if not validate_environment():
        sys.exit(1)

    # Initialize
    sender_email = os.environ["SMVP_USER"]
    email_server = os.environ["SMVP_SERVER"]
    email_token = os.environ["SMVP_TOKEN"]
    receiver_email = args.recipient
    email_subject = args.subject
    email_port = 587

    try:
        with args.file as f:
            text_in = f.read()
    except UnicodeDecodeError:
        msg = f"""
        Unable to process: {args.file.name}
        smvp can only process textfiles (including those with ANSI
        escape sequences) or html files. No email sent.
        """
        print_docstring(msg=msg)
        sys.exit(1)

    # Craft an HTML version compatible with Gmail. If it's not HTML,
    # then filter it through ansi2html to scan for ANSI codes and turn
    # that into HTML. Plaintext will process fine. The text replacement
    # below is to ditch the dull-grey default in ansi2html.
    if not file_is_html(text_in):
        converter = Ansi2HTMLConverter(dark_bg=False)
        html_text = converter.convert(text_in, full=True)
        html_text = html_text.replace("color: #AAAAAA", "color: #FFFFFF")
    else:
        html_text = text_in

    # Set font family and size
    new_style = STYLE.replace("FF", args.font_family)
    new_style = new_style.replace("FS", str(args.font_size))

    soup = BeautifulSoup(html_text, "lxml")
    plain_text = soup.get_text().strip()

    # Gmail strips custom css, so we need to apply inline styles with
    # (!important) to the body tag.
    body_tag = soup.find("body")
    if isinstance(body_tag, Tag):
        body_tag["style"] = new_style

    # Also apply inline styles with (!important) to .ansi2html-content
    # tags
    ansi_content_tags = soup.find_all(class_="ansi2html-content")
    for tag in ansi_content_tags:
        if isinstance(tag, Tag):
            tag["style"] = new_style

    # ! Debug code goes here when testing.

    # Package both parts into a MIME multipart message.
    message = MIMEMultipart("alternative")
    message["Subject"] = email_subject
    message["From"] = sender_email
    message["To"] = receiver_email
    message.attach(MIMEText(plain_text, "plain"))
    message.attach(MIMEText(str(soup), "html"))

    # Create a secure context for the TLS connection
    context = ssl.create_default_context()

    # Send the email
    try:
        server = smtplib.SMTP(email_server, email_port)
        server.starttls(context=context)
        server.login(sender_email, email_token)
        server.sendmail(
            from_addr=sender_email,
            to_addrs=receiver_email,
            msg=message.as_string(),
        )
        print("Message successfully sent.")
    except Exception as e:
        print(e)
    finally:
        server.quit()
    return


# ======================================================================



import sys
from pathlib import Path

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def get_version() -> str:
    """Return the version number of the project.

    Starting at the current directory, search for pyproject.toml,
    working your way up the parents in the directory tree. This allows
    the version number to be maintained on one place.

    IMPORTANT: For this to work properly when you distribute your
    package, you must include the following in your `pyproject.toml`
    file:

    ```
    [tool.hatch.build.targets.wheel.force-include]
    "./pyproject.toml" = "<package name>/pyproject.toml"
    ```

    Returns
    -------
    str
        Version number of the project.
    """
    start = Path(__file__).absolute().parents

    for path in start:
        for file in path.glob("*.toml"):
            if file.name == "pyproject.toml":
                try:
                    with open(file, "rb") as f:
                        tom = tomllib.load(f)
                    return tom["project"]["version"]
                except Exception:
                    return "unknown"

    return "unknown"


# ======================================================================
